<?php

namespace Drupal\edwt\Utility;

/**
 * CssArray is a PHP utility class to convert CSS into an array.
 *
 * Example usage:
 * ```php
 * $css = file_get_contents($file);
 * $cssArray = CssArray::convert($css);
 * print_r($cssArray);
 * ```
 */
class CssArray {

  /**
   * Start to process the CSS string.
   *
   * @param string $css
   *   CSS styles as a string.
   *
   * @return array
   *   Parsed CSS as an associative array.
   */
  public static function convert(string $css): array {
    $css = self::cleanCss($css);
    return self::parseCss($css);
  }

  /**
   * Recursively merges CSS style arrays.
   *
   * @param array $style
   *   CSS style input.
   * @param array &$result
   *   Result array passed by reference.
   */
  private static function setStyleRecursive(array $style, array &$result): void {
    foreach ($style as $name => $val) {
      if (is_array($val)) {
        if (!isset($result[$name])) {
          $result[$name] = [];
        }
        self::setStyleRecursive($val, $result[$name]);
      }
      else {
        $result[$name] = $val;
      }
    }
  }

  /**
   * Parses media CSS blocks.
   *
   * @param string $style
   *   Media CSS block.
   *
   * @return array
   *   Array of parsed media styles.
   */
  private static function processMediaStyle(string $style): array {
    $result = [];
    preg_match_all('/(?ims)([a-z0-9\s\,\.\:#_\-@*()\[\]"=]+)\{([^\}]*)\}(?:\s\})?/', $style, $matches);
    foreach ($matches[0] as $content) {
      self::processContent($content, $result);
    }
    return $result;
  }

  /**
   * Processes a single CSS @media block.
   *
   * @param string $content
   *   The @media block.
   * @param array &$result
   *   The result array.
   */
  private static function processMediaContent(string $content, array &$result): void {
    $tmp = explode('{', $content);
    $mediaClass = explode(',', array_shift($tmp));
    $mediaStyleContent = self::processMediaStyle(substr(implode('{', $tmp), 0, -1));
    foreach ($mediaClass as $mc) {
      $mc = trim(str_replace('@media', '', $mc));
      $tcl = explode(',', trim($mc));
      $r = &$result['@media'];
      foreach ($tcl as $tt) {
        $name = trim($tt);
        if (!isset($r[$name])) {
          $r[$name] = [];
        }
        $r = &$r[$name];
      }
      self::setStyleRecursive($mediaStyleContent, $r);
    }
  }

  /**
   * Breaks down inline styles into a key-value array.
   *
   * @param string $style
   *   CSS declarations block.
   *
   * @return array
   *   Associative array of styles.
   */
  private static function processStyles(string $style): array {
    $result = [];
    $tmp = explode(';', $style);
    foreach ($tmp as $s) {
      $st = explode(':', $s);
      $name = trim(array_shift($st));
      $val = trim(implode(':', $st));
      if ($val !== '') {
        $result[$name] = (substr($val, -1) === '}') ? substr($val, 0, -1) : $val;
      }
    }
    return $result;
  }

  /**
   * Processes a CSS block (with or without @media).
   *
   * @param string $content
   *   The CSS block.
   * @param array &$result
   *   Result passed by reference.
   */
  private static function processContent(string $content, array &$result): void {
    if (strpos($content, '@media') !== false) {
      self::processMediaContent($content, $result);
    }
    else {
      $tmp = explode('{', $content);
      $class = trim($tmp[0]);
      $style = trim($tmp[1]);
      $styleArr = self::processStyles($style);
      $tc = explode(',', $class);
      foreach ($tc as $cl) {
        $tcl = explode(' ', trim($cl));
        $r = &$result;
        foreach ($tcl as $tt) {
          if (!isset($r[trim($tt)])) {
            $r[trim($tt)] = [];
          }
          $r = &$r[trim($tt)];
        }
        $r = $styleArr;
      }
    }
  }

  /**
   * Parses CSS using regex.
   *
   * @param string $css
   *   The raw CSS string.
   *
   * @return array
   *   Parsed CSS array.
   */
  private static function parseCss(string $css): array {
    $result = [];
    preg_match_all('/(?ims)([a-z0-9\s\,\.\:#_\-@*()\[\]"=]+)\{([^\}]*)\}(?:\s\})?/', $css, $matches);
    $countBraces = 0;
    $multiContent = '';
    foreach ($matches[0] as $content) {
      $countBraces += substr_count($content, '{');
      $countBraces -= substr_count($content, '}');
      if ($countBraces === 0) {
        if (!empty($multiContent)) {
          $multiContent .= $content;
          self::processContent($multiContent, $result);
          $multiContent = '';
        }
        else {
          self::processContent($content, $result);
        }
      }
      else {
        $multiContent .= $content;
      }
    }
    ksort($result);
    return $result;
  }

  /**
   * Remove space from numeric values (callback).
   *
   * @param array $matches
   *   Regex matches.
   *
   * @return string
   *   Modified string.
   */
  private static function removeSpace(array $matches): string {
    return str_replace(' ', '', $matches[0]);
  }

  /**
   * Cleans the raw CSS string to prepare it for parsing.
   *
   * @param string $css
   *   Raw CSS.
   *
   * @return string
   *   Cleaned CSS.
   */
  private static function cleanCss(string $css): string {
    $css = preg_replace('/\@font\-face\s?\{[^}]*\}/m', '', $css);
    $css = preg_replace('/\@import\surl\([^}]*\)\;/m', '', $css);
    $css = preg_replace('/(data\:[^;]+);/i', '$1#&', $css);
    $css = preg_replace('!/\*.*?\*/!s', '', $css);
    $css = preg_replace('/\n\s*\n/', "\n", $css);
    $css = preg_replace('/\>/m', ' ', $css);

    $find = [
      '::before', ':before', '::after', ':after', ':active', ':link', ':visited',
      ':first-child', ':last-child', ':nth-child', ':hover', ':focus',
      '._', '.', '#', '@media screen and', ') and (', '[',
    ];
    $replace = [
      ' :before', ' :before', ' :after', ' :after', ' :active', ' :link', ' :visited',
      ' :first-child', ' :last-child', ' :nth-child', ' :hover', ' :focus',
      ' ._', ' .', ' #', '@media', '), @media (', ' [',
    ];

    $css = str_replace($find, $replace, $css);
    $css = trim(preg_replace_callback('/\d\s\.\d/m', [self::class, 'removeSpace'], $css));
    $css = trim(preg_replace('/\s+/', ' ', $css));

    return $css;
  }

}
